# rdl_compute.r

# Change the following line to point to your current working directory
working_directory <- "C:\\Users\\UserName\\Documents\\QED"
set.seed(0)

# Specify positive and negative compound sets as CSV files: used to train RDL
pos_set_filename <- "Drugs.csv"
neg_set_filename <- "ChEMBL09_QED.csv"

# CSV file containing compounds for which we wish to compute the RDL
input_filename <- "DrugBankQED.csv"           

# CSV file to which to output computed RDLs
output_filename <- "rdls.csv"

# Properties to use
prop_names <- c("MW","ALOGP","HBA","HBD","PSA","ROTB","AROM","ALERTS")

# Whether to only use a sample of the negative set to train RDL, and if so, how big the sample should be
use_neg_sample <- TRUE
neg_sample_size <- 1000

###############################################################################################################
setwd(working_directory)
pos_set <- read.csv(file = pos_set_filename, head = TRUE, sep = ",")    

full_neg_set <- read.csv(file = neg_set_filename, head = TRUE, sep = ",")    
neg_set <- NULL
if (use_neg_sample == TRUE) {
    inds <- 1:dim(full_neg_set)[1]
    sample_inds <- sample(inds, neg_sample_size)
    neg_set <- full_neg_set[sample_inds,]       
} else {
    neg_set <- full_neg_set
}                           

compound_set <- read.csv(file = input_filename, head = TRUE, sep = ",")

#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
# Histogram bin width optimisation
# 2006 Author Hideaki Shimazaki
# Department of Physics, Kyoto University
# shimazaki at ton.scphys.kyoto-u.ac.jp
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
sshist <- function(x,N)
{
	C <- numeric(length(N))
	D <- C

	for (i in 1:length(N)) {
		D[i] <- diff(range(x))/N[i]

		edges = seq(min(x),max(x),length=N[i])
		hp <- hist(x, breaks = edges, plot = FALSE)
		ki <- hp$counts

		k <- mean(ki)
		v <- sum((ki-k)^2)/N[i]

		C[i] <- (2*k-v)/D[i]^2	                        # cost function
	}

	idx <- which.min(C)
	optD <- D[idx]

	edges <- seq(min(x),max(x),length=N[idx])
	h = hist(x, breaks = edges, plot = FALSE)

	return(h)
}

hists <- NULL
classbdries <- NULL
proportions_pos_set <- NULL
proportions_neg_set <- NULL
rel_likelihood <- NULL
ratio_models <- NULL
compound_rdl_components <- NULL
compound_rdls <- rep(1, dim(compound_set)[1])
df_rdls <- data.frame(nrow = length(compound_rdls))

for (prop_name in prop_names)
{
    df_rdls <- cbind(df_rdls, compound_set[[prop_name]])
}

# Compute the relative likelihood for each property
for (prop_name in prop_names)
{
        propvals_pos_set <- eval(as.name(prop_name), pos_set)
        propvals_neg_set <- eval(as.name(prop_name), neg_set)
        
        if (prop_name == "MW" ||
            prop_name == "ALOGP")
        {        
            hists[[prop_name]] <- sshist(propvals_pos_set, 20:100)
            #hists[[prop_name]] <- hist(propvals_pos_set, breaks = 6, plot = FALSE)
        }
        else if (prop_name == "PSA")
        {
            hists[[prop_name]] <- sshist(propvals_pos_set, 15:30)
        }
        else if (prop_name == "HBA" || prop_name == "HBD" || prop_name == "ROTB" ||
                 prop_name == "AROM" || prop_name == "ALERTS")
        {
            hists[[prop_name]] <- hist(propvals_pos_set, seq(min(propvals_pos_set)-0.5, max(propvals_pos_set)+0.5, by=1), plot = FALSE)
        }        
        
        new_hists <- NULL
        
        
        classbdries[[prop_name]] <- hists[[prop_name]]$breaks
        classcounts[[prop_name]] <- hists[[prop_name]]$counts
        
        if (prop_name == "MW" || prop_name == "ALOGP" || prop_name == "PSA" || prop_name == "HBA" || prop_name == "HBD" || prop_name == "ROTB" ||
            prop_name == "AROM" || prop_name == "ALERTS")
        {
            for (i in 1:length(classcounts[[prop_name]]))
            {
                j <- i
                while (j <= length(classcounts[[prop_name]]) && classcounts[[prop_name]][j] < 5)
                {
                    if (j != 1 && j <= length(classcounts[[prop_name]]) - 2)
                    {
                        classcounts[[prop_name]] <- c(classcounts[[prop_name]][1:(j-1)], classcounts[[prop_name]][j] + classcounts[[prop_name]][j+1], classcounts[[prop_name]][(j+2):length(classcounts[[prop_name]])])
                        classbdries[[prop_name]] <- c(classbdries[[prop_name]][1:j], classbdries[[prop_name]][(j+2):length(classbdries[[prop_name]])])
                    }
                    else if (j == 1)
                    {
                        classcounts[[prop_name]] <- c(classcounts[[prop_name]][1] + classcounts[[prop_name]][2], classcounts[[prop_name]][3:length(classcounts[[prop_name]])])
                        classbdries[[prop_name]] <- c(classbdries[[prop_name]][1], classbdries[[prop_name]][3:length(classbdries[[prop_name]])])
                    }
                    else if (j == length(classcounts[[prop_name]]) - 1)
                    {
                        classcounts[[prop_name]] <- c(classcounts[[prop_name]][1:(j-1)], classcounts[[prop_name]][j] + classcounts[[prop_name]][j+1])
                        classbdries[[prop_name]] <- c(classbdries[[prop_name]][1:j], classbdries[[prop_name]][j+2])                    
                    }
                    else if (j == length(classcounts[[prop_name]]))
                    {
                        classcounts[[prop_name]] <- c(classcounts[[prop_name]][1:(j-2)], classcounts[[prop_name]][j-1] + classcounts[[prop_name]][j])
                        classbdries[[prop_name]] <- c(classbdries[[prop_name]][1:(j-1)], classbdries[[prop_name]][j+1])                    
                    }                    
                }
            }
            
            hists[[prop_name]] <- hist(propvals_pos_set, breaks = classbdries[[prop_name]], plot = FALSE)
        }
        
        # Compute the proportion of compounds in each bin for the positive set
        proportions_pos_set[[prop_name]] <- hists[[prop_name]]$counts / length(propvals_pos_set)           

        # Compute the proportion of compounds in each bin for the negative set
        bincounts <- NULL                     # number of compounds in each bin for the negative set
        for (i in 1:(length(classbdries[[prop_name]]) - 1))
        {
            bincounts <- c(bincounts, length(which(classbdries[[prop_name]][i] <= propvals_neg_set &
                                                   propvals_neg_set <= classbdries[[prop_name]][i + 1])))
        }
        proportions_neg_set[[prop_name]] <- bincounts / length(propvals_neg_set)

        # Compute relative likelihood for each bin
        rel_likelihood[[prop_name]] <- proportions_pos_set[[prop_name]] / proportions_neg_set[[prop_name]]
        rel_likelihood[[prop_name]][rel_likelihood[[prop_name]] == Inf] <- 1.0

        ratio_models[[prop_name]] <- loess(rel_likelihood[[prop_name]] ~ hists[[prop_name]]$mids, control=loess.control(surface="direct"))
                
        compound_rdl_components[[prop_name]] <- predict(ratio_models[[prop_name]], compound_set[[prop_name]])
        compound_rdl_components[[prop_name]] <- ifelse(compound_rdl_components[[prop_name]] > 0, compound_rdl_components[[prop_name]], 0)
        compound_rdls <- compound_rdls * compound_rdl_components[[prop_name]]
        df_rdls <- cbind(df_rdls, compound_rdl_components[[prop_name]])
}

compound_rdls <- compound_rdls ^ (1.0 / length(prop_names))                 # geometric mean over each RDL component

# Write out RDL for each compound to CSV file
df_rdls <- cbind(df_rdls, compound_rdls)
df_rdls[1] <- NULL
colnames(df_rdls) <- c(prop_names, paste(prop_names, "_RDL"), "RDL")
write.csv(df_rdls, output_filename, row.names=FALSE)   











